// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'document.dart';
import 'index_directory.dart';

class IndexWriter {

  final IndexDirectory _directory;

  IndexWriter(this._directory);

  void write(Iterable<Document> documents) {
    int docId = 0;
    var indexes = <String,InvertedIndexBuilder>{};
    var data = <String,StoredDataBuilder>{};
    for (Document doc in documents) {
      _updateDocument(docId, doc, indexes, data);
      docId++;
    }
    _storeIndexes(indexes);
    _storeData(data);
  }

  void _updateDocument(int docId, Document document,
      Map<String,InvertedIndexBuilder> indexes,
      Map<String,StoredDataBuilder> data) {
    for (Field field in document.fields) {
      String name = field.name;
      Iterable<String> tokens = field.analyzer.getTokens(field.text);
      if (!indexes.containsKey(name)) {
        indexes[name] = new InvertedIndexBuilder().append(tokens, docId);
      } else {
        indexes[name].append(tokens, docId);
      }
      if (!data.containsKey(name)) {
        data[name] = new StoredDataBuilder().append(docId, field.text);
      } else {
        data[name].append(docId, field.text);
      }
    }
  }

  void _storeIndexes(Map<String,InvertedIndexBuilder> indexes) {
    for (String name in indexes.keys) {
      _directory.outputIndex(name, indexes[name].build());
    }
  }

  void _storeData(Map<String,StoredDataBuilder> data) {
    for (String name in data.keys) {
       _directory.outputData(name, data[name].build());
    }
  }

}

class InvertedIndexBuilder {

  var _invertedIndex = <String,List<int>>{};

  InvertedIndexBuilder append(Iterable<String> tokens, int docId) {
    for (String token in tokens) {
      if(!_invertedIndex.containsKey(token)) {
        var list = <int>[]..add(docId);
        _invertedIndex[token] = list;
      } else {
        var list = _invertedIndex[token];
        if (!list.contains(docId)) {
          list.add(docId);
        }
      }
    }
    return this;
  }

  Map<String,List<int>> build() {
    var ret = {};
    List<String> terms = _invertedIndex.keys.toList();
    terms.sort();
    for (String term in terms) {
      ret[term] = _invertedIndex[term];
    }
    return ret;
  }
}

class StoredDataBuilder {

  var _storedData = <int,String>{};

  StoredDataBuilder append(int docId, String text) {
    _storedData[docId] = text;
    return this;
  }

  Map<int,String> build() {
    return _storedData;
  }

}